@@ -1,6 +1,3 @@
-# Code to translate from a reduced AST into C++ code. This is done by traversing the
-# Python AST code.
-
 import ast
 import logging
 from abc import ABC, abstractmethod
@@ -89,7 +86,7 @@ def get_ttree_type(rep):
         return rep.cpp_type()
 
 
-def determine_type_mf(parent_type, function_name):
+def determine_type_mf(parent_type: ctyp.terminal, function_name: str) -> ctyp.MethodInvokeInfo:
     '''
     Determine the return type of the member function. Do our best to make
     an intelligent case when we can.
@@ -102,19 +99,19 @@ def determine_type_mf(parent_type, function_name):
         raise RuntimeError("Internal Error: Trying to call member function for a type we do not know!")
     # If we are doing one of the normal "terminals", then we can just bomb. This should not happen!
 
-    rtn_type = ctyp.method_type_info(str(parent_type), function_name)
+    t_parent = parent_type.type
+    rtn_type = ctyp.method_type_info(t_parent, function_name)
     if rtn_type is not None:
         return rtn_type
 
     # We didn't know it. Lets make a guess, and error out if we are clearly making a mistake.
     base_types = ['double', 'float', 'int']
-    s_parent_type = str(parent_type)
-    if s_parent_type in base_types:
+    if t_parent in base_types:
         raise xAODTranslationError(f'Unable to call method {function_name} on type {str(parent_type)}.')
 
     # Ok - we give up. Return a double.
-    logging.getLogger(__name__).warning(f"Warning: assumping that the method '{str(s_parent_type)}.{function_name}(...)' has return type 'double'. Use cpp_types.add_method_type_info to suppress (or correct) this warning.")
-    return ctyp.terminal('double')
+    logging.getLogger(__name__).warning(f"Warning: assuming that the method '{str(t_parent)}::{function_name}(...)' has return type 'double'. Use cpp_types.add_method_type_info to suppress (or correct) this warning.")
+    return ctyp.MethodInvokeInfo(ctyp.terminal('double'), 0)
 
 
 def _extract_column_names(names_ast: ast.AST) -> List[str]:
@@ -130,7 +127,7 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
     Drive the conversion to C++ from the top level query
     """
 
-    def __init__(self, prefix, is_loop_var_a_ref):
+    def __init__(self, prefix):
         r'''
         Initialize the visitor.
         '''
@@ -138,11 +135,13 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         self._gc = generated_code()
         self._arg_stack = argument_stack()
         self._prefix = prefix
-        self._is_loop_var_a_ref = is_loop_var_a_ref
 
     def include_files(self):
         return self._gc.include_files()
 
+    def link_libraries(self) -> List[str]:
+        return self._gc.link_libraries()
+
     def emit_query(self, e):
         'Emit the parsed lines'
         self._gc.emit_query_code(e)
@@ -163,13 +162,14 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         '''
         rep = getattr(node, 'rep', None)
         if rep is not None:
-            if not self._gc.current_scope().starts_with(rep.scope()):
+            rep_scope = getattr(node, 'scope', rep.scope())
+            if not self._gc.current_scope().starts_with(rep_scope):
                 rep = None
 
         if rep is None:
             FuncADLNodeVisitor.visit(self, node)
 
-    def get_rep(self, node: ast.AST, retain_scope: bool = False) -> Union[crep.cpp_rep_base]:
+    def get_rep(self, node: ast.AST, retain_scope: bool = False) -> crep.cpp_rep_base:
         r'''Return the rep for the node. If it isn't set yet, then run our visit on it.
 
         node - The ast node to generate a representation for.
@@ -288,16 +288,21 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         'Look up the in our local dict. This takes care of function arguments, etc.'
         return self._arg_stack.lookup_name(id)
 
-    def make_sequence_from_collection(self, rep):
+    def make_sequence_from_collection(self, rep: crep.cpp_collection) -> crep.cpp_sequence:
         '''
         Take a collection and produce a sequence. Eventually this should likely be some sort of
         plug-in architecture. But for now, we will just assume everything looks like a vector. When
         it comes time for a new type, this is where it should go.
         '''
-        element_type = rep.cpp_type().element_type()
-        element_type = element_type.get_dereferenced_type() if self._is_loop_var_a_ref else element_type
+        cpp_type = rep.cpp_type()
+        assert isinstance(cpp_type, ctyp.collection)
+        element_type = cpp_type.element_type
         iterator_value = crep.cpp_value(unique_name("i_obj"), None, element_type)  # type: ignore
-        l_statement = statement.loop(iterator_value, crep.dereference_var(rep), is_loop_var_a_ref=self._is_loop_var_a_ref)  # type: ignore
+
+        # It could be this should deref until p_depth is 0
+        collection = crep.dereference_var(rep)
+
+        l_statement = statement.loop(iterator_value, collection)
         self._gc.add_statement(l_statement)
         iterator_value.reset_scope(self._gc.current_scope())
 
@@ -336,7 +341,7 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
             return r
 
         # If it isn't a sequence or a collection, then something has gone wrong.
-        raise Exception(f"Unable to generate a sequence from the given AST. Either there is an internal error, or you are trying to manipulate a '{type(rep).__name__}' as a sequence (ast is: {ast.dump(generation_ast)})")
+        raise Exception(f"Unable to generate a sequence from the given AST. Either there is an internal error, or you are trying to manipulate a {str(rep)} ('{type(rep).__name__}') as a sequence (ast is: {ast.dump(generation_ast)})")
 
     def visit_Call_Lambda(self, call_node):
         'Call to a lambda function. We propagate the arguments through the function'
@@ -563,15 +568,20 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         function_name = call_node.func.attr  # type: ignore
         if not isinstance(calling_against, crep.cpp_value):
             # We didn't use get_rep_value above because now we can make a better error message.
-            raise Exception("Do not know how to call '{0}' on '{1}'".format(function_name, type(calling_against).__name__))
+            raise ValueError("Do not know how to call '{0}' on '{1}'".format(function_name, type(calling_against).__name__))
 
         # We support member calls that directly translate only. Here, for example, this is only for
         # obj.pt() or similar. The translation is direct.
-        c_stub = calling_against.as_cpp() + ("->" if calling_against.is_pointer() else ".")
-        result_type = determine_type_mf(calling_against.cpp_type(), function_name)
+        m_info = determine_type_mf(calling_against.cpp_type(), function_name)
+        c_stub = crep.base_type_member_access(calling_against, m_info.deref_depth)
 
+        # Support returned collections or values depending on the result type.
         args = call_node.args
-        crep.set_rep(call_node, crep.cpp_value(c_stub + function_name + f"({','.join(self.get_rep(arg).as_cpp() for arg in args)})", calling_against.scope(), result_type))  # type: ignore
+        v_name = f"{c_stub}{function_name}({','.join(self.get_rep(arg).as_cpp() for arg in args)})"  # type: ignore
+        if isinstance(m_info.r_type, ctyp.collection):
+            crep.set_rep(call_node, crep.cpp_collection(v_name, calling_against.scope(), m_info.r_type))
+        else:
+            crep.set_rep(call_node, crep.cpp_value(v_name, calling_against.scope(), m_info.r_type))
 
     def visit_function_ast(self, call_node):
         'Drop-in replacement for a function'
@@ -580,7 +590,9 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         arg_reps = [self.get_rep_value(a) for a in call_node.args]
 
         # Code up a call
-        r = crep.cpp_value('{0}({1})'.format(cpp_func.cpp_name, ','.join(a.as_cpp() for a in arg_reps)), self._gc.current_scope(), cpp_type=cpp_func.cpp_return_type)
+        r = crep.cpp_value(f'{cpp_func.cpp_name}({",".join(a.as_cpp() for a in arg_reps)})',
+                           self._gc.current_scope(),
+                           cpp_type=ctyp.terminal(cpp_func.cpp_return_type))
 
         # Include files and return the resulting expression
         for i in cpp_func.include_files:
@@ -620,8 +632,8 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         variable = node.attr
         if not isinstance(obj, crep.cpp_value):
             raise Exception("Do not know how to get member '{0}' of '{1}'".format(variable, type(obj).__name__))
-        result_type = determine_type_mf(obj.cpp_type(), variable)
-        crep.set_rep(node, crep.cpp_value(f"{obj.as_cpp()}{'->' if obj.is_pointer() else '.'}{variable}", self._gc.current_scope(), result_type))
+        m_info = determine_type_mf(obj.cpp_type(), variable)
+        crep.set_rep(node, crep.cpp_value(f"{crep.base_type_member_access(obj, m_info.deref_depth)}{variable}", self._gc.current_scope(), m_info.r_type))
 
     def visit_Name(self, name_node: ast.Name):
         'Visiting a name - which should represent something'
@@ -636,7 +648,7 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
             raise Exception("Do not know how to take the index of type '{0}'".format(v.cpp_type()))  # type: ignore
 
         index = self.get_rep(node.slice)
-        crep.set_rep(node, crep.cpp_value(f"{v.as_cpp()}{'->' if v.is_pointer() else '.'}at({index.as_cpp()})", self._gc.current_scope(), cpp_type=v.get_element_type()))  # type: ignore
+        crep.set_rep(node, crep.cpp_value(f"{crep.base_type_member_access(v)}at({index.as_cpp()})", self._gc.current_scope(), cpp_type=v.get_element_type()))  # type: ignore
 
     def visit_Index(self, node):
         'We can only do single items, we cannot do slices yet'
@@ -773,7 +785,7 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         '''
 
         # The result of this test
-        result = crep.cpp_variable(unique_name('bool_op'), self._gc.current_scope(), cpp_type='bool')
+        result = crep.cpp_variable(unique_name('bool_op'), self._gc.current_scope(), cpp_type=ctyp.terminal('bool'))
         self._gc.declare_variable(result)
 
         # How we check and short-circuit depends on if we are doing and or or.
@@ -797,13 +809,13 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         crep.set_rep(node, result)
 
     def visit_Num(self, node):
-        crep.set_rep(node, crep.cpp_value(node.n, self._gc.current_scope(), guess_type_from_number(node.n)))
+        crep.set_rep(node, crep.cpp_value(str(node.n), self._gc.current_scope(), guess_type_from_number(node.n)))
 
     def visit_Str(self, node):
         crep.set_rep(node, crep.cpp_value('"{0}"'.format(node.s), self._gc.current_scope(), ctyp.terminal("string")))
 
     def visit_Constant(self, node: ast.Constant):
-        '''Visit Constnat node.
+        '''Visit constant node.
 
         Note that this has to take the place of `visit_Str` and `visit_Num` as we roll
         python versions forward. So those methods and this method must be kept in sync.
@@ -815,11 +827,23 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         if type(value) is str:
             crep.set_rep(node, crep.cpp_value(f'"{value}"', self._gc.current_scope(), ctyp.terminal("string")))
         elif type(value) is int:
-            crep.set_rep(node, crep.cpp_value(value, self._gc.current_scope(), ctyp.terminal("int")))
+            crep.set_rep(node, crep.cpp_value(str(value), self._gc.current_scope(), ctyp.terminal("int")))
         elif type(value) is float:
-            crep.set_rep(node, crep.cpp_value(value, self._gc.current_scope(), ctyp.terminal("double")))
+            crep.set_rep(node, crep.cpp_value(str(value), self._gc.current_scope(), ctyp.terminal("double")))
+        elif type(value) is bool:
+            cpp_value = "true" if value else "false"
+            crep.set_rep(node, crep.cpp_value(cpp_value, self._gc.current_scope(), ctyp.terminal('bool')))
         else:
-            raise Exception(f"Unsupported constant type: {type(value)}")
+            raise ValueError(f"Unsupported constant type: {type(value)}")
+
+    def visit_NameConstant(self, node: ast.NameConstant):
+        value = node.value
+        if value is True:
+            crep.set_rep(node, crep.cpp_value("true", self._gc.current_scope(), ctyp.terminal('bool')))
+        elif value is False:
+            crep.set_rep(node, crep.cpp_value("false", self._gc.current_scope(), ctyp.terminal('bool')))
+        else:
+            raise ValueError(f"Unsupported constant: {value}")
 
     def code_fill_ttree(self, e_rep: crep.cpp_rep_base, e_name: crep.cpp_variable,
                         scope_fill: Union[gc_scope, gc_scope_top_level]) -> Union[gc_scope, gc_scope_top_level]:
@@ -964,14 +988,14 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
 
         assert len(args) == 2
         source = args[0]
-        selection = cast(ast.Lambda, args[1])
+        selection = args[1]
+        assert isinstance(selection, ast.Lambda)
 
         # Make sure we are in a loop
         seq = self.as_sequence(source)
 
         # Simulate this as a "call"
-        selection = lambda_unwrap(selection)
-        c = ast.Call(func=selection, args=[seq.sequence_value().as_ast()])
+        c = ast.Call(func=lambda_unwrap(selection), args=[seq.sequence_value().as_ast()])
         new_sequence_value = cast(crep.cpp_value, self.get_rep(c))
 
         # We need to build a new sequence.
@@ -993,12 +1017,11 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         # Make sure the source is around. We have to do this because code generation in this
         # framework is lazy. And if the `selection` function does not use the source, and
         # looking at that source might generate a loop, that loop won't be generated! Ops!
-        _ = self.as_sequence(source)
+        seq = self.as_sequence(source)
 
         # We need to "call" the source with the function. So build up a new
         # call, and then visit it.
-
-        c = ast.Call(func=lambda_unwrap(selection), args=[source])
+        c = ast.Call(func=lambda_unwrap(selection), args=[seq.sequence_value().as_ast()])
 
         # Get the collection, and then generate the loop over it.
         # It could be that this comes back from something that is already iterating (like a Select statement),
@@ -1083,6 +1106,7 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         source = args[0]
 
         # Make sure we are in a loop.
+        cs = self._gc.current_scope()
         seq = self.as_sequence(source)
 
         # The First terminal works by protecting the code with a if (first_time) {} block.
@@ -1115,4 +1139,4 @@ class query_ast_visitor(FuncADLNodeVisitor, ABC):
         # Otherwise return a new version of the value.
         first_value = sv if isinstance(sv, crep.cpp_sequence) else sv.copy_with_new_scope(self._gc.current_scope())
 
-        crep.set_rep(node, first_value)
+        crep.set_rep(node, first_value, cs)
